'use strict'
import 'dotenv/config';

import { v4 as uuid } from 'uuid';
import { join as pathJoin, basename } from 'path';
import { readFile, writeFile, access } from 'fs/promises'; 
import{ createReadStream } from 'fs';
import { createHash } from 'crypto';
import * as mime from 'mime-types';
import { Endpoint, S3 } from 'aws-sdk';
import { tmpdir } from 'os';
import s3Stream = require('s3-upload-stream');

import { createLog } from '../log';
import type { Logger } from 'winston';
import { envString } from '../env';

const TMP_DIR = (envString('FILE3_DIR', null) !== null) ? envString('FILE3_DIR', '/tmp') : envString('TMP_DIR', tmpdir());

interface FileRecord {
	id : string;
	created : number;
	name : string;
	public : string;
	path : string;
	path_hash : string;
	hash : string;
	type : string;
	size : number;
}

export class Files3 {
    private writeable : boolean = false;
    private s3 : S3;
    private s3Stream : any;
    private bucket : string;
    private endpoint : string;
    private log : Logger;

    /**
     * @constructor
     * 
     */

    constructor (bucket : string, writeable : boolean = true) {
		const S3_ENDPOINT : string = envString('S3_ENDPOINT', 'http://127.0.0.1:9000');
		const spacesEndpoint : Endpoint = new Endpoint(S3_ENDPOINT);
        const s3Config = {
            accessKeyId:  envString('S3_ACCESS_KEY', 'YOUR-ACCESSKEYID'),
            secretAccessKey:  envString('S3_ACCESS_SECRET', 'YOUR-SECRETACCESSKEY'),
            endpoint: spacesEndpoint as unknown as string,
            signatureVersion: 'v4'
        };
		this.endpoint = S3_ENDPOINT;
		this.s3  = new S3(s3Config);
        this.s3Stream = s3Stream(this.s3);
        this.log = createLog('files3');

		this.bucket = bucket;
        this.writeable = writeable;
    }

    /**
     * Create a SHA256 hash of any data provided.
     **/
    private hash (data : any) {
        return createHash('sha256').update(data).digest('base64');
    }

    /**
     * Check if file exists
     **/
    private async exists (path : string) : Promise<boolean> {  
        try {
            await access(path);
            return true;
        } catch {
            return false;
        }
    }

    /**
     * Read file from disk as buffer and create hash of the data.
     **/
    private async hashFile (filePath : string) {
    	let data;
    	try {
    		data = await readFile(filePath);
    	} catch (err) {
    		this.log.error(err);
    	}
    	return this.hash(data);
    }
    /**
     * create a file object on an S3 bucket and upload data.
     * Reads into memory
     **/
    public async create (file : any, keyName? : string) {
        if (!this.writeable) return false;
        const id : string = await this.hashFile(file);
		const ext : string | false  = mime.extension(file.mimetype);
		const key : string = typeof keyName !== 'undefined' ? keyName : `${id}.${ext}`;
		const webPath : string = pathJoin('/files/', this.bucket, key);
		const publicPath : string = pathJoin(`${this.bucket}.${this.endpoint}`, key);
		const record : FileRecord = {
			id,
			created : +new Date(),
			name : file.originalname,
			public : publicPath,
			path : webPath,
			path_hash : this.hash(webPath),
			hash : null,
			type : file.mimetype,
			size : null
		};
		const params : S3.PutObjectRequest = {
			Bucket: this.bucket, 
			Key: key, 
			Body: null
		};

		params.Body = file.buffer;
		record.hash = this.hash(file.buffer);
		record.size = file.buffer.byteLength;

		return new Promise((resolve : Function, reject : Function) => {
			return this.s3.putObject(params, function (err : Error, data : any) {
				if (err) {
					this.log.error(err)
					return reject(err)
				} else  { 
					this.log.info(`Saved file ${record.path}`);
					return resolve(record);
				}
			});
		});
    }

    /*
     *  Create an s3 record using only a path reference of a local file
     *  Reads into memory
     */

    public async createFromPath (filePath : string, keyName? : string) {
        const filename : string = basename(filePath);
        const mimetype : string | false = mime.lookup(filePath);
        const key = typeof keyName !== 'undefined' ? keyName : null;
        let file : any;
        let buffer : any;

        try {
            buffer = await readFile(filePath);
        } catch (err) {
            this.log.error('createFromPath', err);
        }

        file = {
            buffer,
            mimetype,
            originalname : filename
        };
        
        return this.create(file, key);
    }

    /**
     *  Create an s3 record from a stream
     *  Pass "createReadStream('')" object
     **/

    /*public async createStream (file : any) {
        if (!this.writeable) return false;
        const id : string = uuid();
        const ext : string | false = mime.extension(file.mimetype);
        const key : string = `${id}.${ext}`;
        const webPath : string = pathJoin('/files/', this.bucket, key);
        const publicPath : string = pathJoin(`${this.bucket}.${this.endpoint}`, key);
        const record : FileRecord = {
            id,
            created : +new Date(),
            name : file.originalname,
            public : publicPath,
            path : webPath,
            path_hash : this.hash(webPath),
            hash : null,
            type : file.mimetype,
            size : null
        };
        const params : S3Params = {
            Bucket: this.bucket, 
            Key: key
        };
        const upload = this.s3Stream.upload(params);

        upload.maxPartSize(20971520); // 20 MB
        upload.concurrentParts(5);

        return new Promise((resolve : Function, reject : Function) => {
            upload.on('error', (err : Error) => {
                return reject(err);
            });
            upload.on('part', (details : any) => {
                this.log.info(`${details.ETag} - part: ${details.PartNumber} received: ${details.receivedSize} uploaded: ${details.uploadedSize}`)
            });
            upload.on('uploaded', (details : any) => {
                record.hash = details.ETag;
                record.size = details.uploadedSize;
                this.log.info(`Saved file ${record.path}`);
                return resolve(record);
            });
            this.log.info(`Streaming ${record.path} to S3`);
            stream.pipe(upload);
        });
    }*/

    /**
     * Create a stream . Bind to busboy.on('file', files3.createStream)
     * ex. (express POST route callback)
     * var busboy = new Busboy({ headers: req.headers });
     * busboy.on('file', files3.createStream)
     * req.pipe(busboy);
     **/

   /* public async createStreamExpress (fieldname : string, file : any, filename : string, encoding : any, mimetype : string) {
        if (!this.writeable) return false;
        const id : string = uuid();
        const ext : string | false = mime.extension(mimetype);
        const key : string = `${id}.${ext}`;
        const webPath : string = pathJoin('/files/', this.bucket, key);
        const publicPath : string = pathJoin(`${this.bucket}.${this.endpoint}`, key);
        const record : FileRecord = {
            id,
            created : +new Date(),
            name : filename,
            public : publicPath,
            path : webPath,
            path_hash : this.hash(webPath),
            hash : null,
            type : file.mimetype,
            size : null
        };
        const params : S3Params = {
            Bucket: this.bucket, 
            Key: key
        };
        const upload = this.s3Stream.upload(params);

        upload.maxPartSize(20971520); // 20 MB
        upload.concurrentParts(5);

        return new Promise((resolve : Function, reject : Function) => {
            var s3 = new AWS.S3({
         params: {Bucket: 'sswa', Key: filename, Body: file},
         options: {partSize: 5 * 1024 * 1024, queueSize: 10}   // 5 MB
      });
      s3.upload().on('httpUploadProgress', function (evt) {
         this.log.info(evt);
      }).send(function (err, data) {
         s3UploadFinishTime = new Date();
         if(busboyFinishTime && s3UploadFinishTime) {
            res.json({
               uploadStartTime: uploadStartTime,
               busboyFinishTime: busboyFinishTime,
               s3UploadFinishTime: s3UploadFinishTime
            });
         }
         this.log.info(err, data);
      });
            file.on('data', ( data : any ) => {
                upload.on('error', (err : Error) => {
                    return reject(err);
                });
                upload.on('part', (details : any) => {
                    this.log.info(`${details.ETag} - part: ${details.PartNumber} received: ${details.receivedSize} uploaded: ${details.uploadedSize}`)
                });
                upload.on('uploaded', (details : any) => {
                    record.hash = details.ETag;
                    record.size = details.uploadedSize;
                    this.log.info(`Saved file ${record.path}`)
                    return resolve(record)
                });
                data.pipe(upload);
            })
        });
    }*/

    /**
     * Create a stream from a path on the local device
     *
     * @param {string} filePath     Path to file
     * @param {string} keyName     (optional) Predefined key
     **/

    public async createStreamFromPath (filePath : string, keyName?: string) {
        if (!this.writeable) return false;
        const id : string = uuid();
        const fileName : string = basename(filePath);
        const mimetype : string | false = mime.lookup(filePath);
        const ext : string | false = mime.extension(fileName);
        const key : string = typeof keyName !== 'undefined' ? keyName : `${id}.${ext}`;
        const webPath : string = pathJoin('/files/', this.bucket, key);
        const publicPath : string = pathJoin(`${this.bucket}.${this.endpoint}`, key);
        const record : FileRecord = {
            id,
            created : +new Date(),
            name : fileName,
            public : publicPath,
            path : webPath,
            path_hash : this.hash(webPath),
            hash : null,
            type : mimetype ? mimetype : null,
            size : null
        };
        const params : S3.PutObjectRequest = {
            Bucket: this.bucket, 
            Key: key
        };
        const upload = this.s3Stream.upload(params);
        const stream = createReadStream(filePath);

        upload.maxPartSize(20971520); // 20 MB
        upload.concurrentParts(5);

        return new Promise((resolve : Function, reject : Function) => {
            upload.on('error', (err : Error) => {
                return reject(err);
            });
            upload.on('part', (details : any) => {
                this.log.info(`${details.ETag} - part: ${details.PartNumber} received: ${details.receivedSize} uploaded: ${details.uploadedSize}`)
            });
            upload.on('uploaded', (details : any) => {
                record.hash = details.ETag;
                record.size = details.uploadedSize;
                this.log.info(`Saved file ${record.path}`);
                return resolve(record);
            });
            this.log.info(`Streaming ${record.path} to S3`);
            stream.pipe(upload);
        });
    }

    /**
     * Read a file from S3 using a key
     *
     * @param {string} key File key
     *
     * @returns {string} File data
     **/

    public async read (key : string) {
    	const params : S3.PutObjectRequest = {
    		Bucket: this.bucket, 
    		Key: key
    	};
    	return new Promise((resolve : Function, reject : Function) => {
    		return this.s3.getObject(params, (err : Error, data : any) => {
    			if (err) {
    				return reject(err);
    			}
    			return resolve(data.Body) //buffer
    		});
    	});
    }

    /*
        const xhr = new XMLHttpRequest();
        xhr.open('PUT', signedUrl);
        xhr.setRequestHeader('Content-Type', file.type);
        xhr.setRequestHeader('x-amz-acl', 'public-read');
        xhr.send(file);
    */

    /**
     * Get a signed put key for writing
     *
     * @param {string} key          Key that file will be located at
     * @param {string} fileType     Mimetype of file
     *
     * @returns {string} Url of signed key
     **/
    public async signedPutKey (key : string, fileType : string) {
        const s3Params : S3.PutObjectRequest = {
            Bucket: this.bucket,
            Key: key,
            ContentType: fileType,
            Expires: new Date((new Date()).getTime() + 24 * 60 * 60 * 1000) //1 day
            //ACL: 'public-read',
            
        };
        return new Promise((resolve : Function, reject : Function) => {
            return this.s3.getSignedUrl('putObject', s3Params, (err : Error, url : string) => {
                if (err) {
                    return reject(err);
                }
                return resolve(url);
            });
        });
    }

    /**
     * Get a signed read key for writing
     *
     * @param {string} key          Key that file will be located at
     *
     * @returns {string} Url of signed key
     **/
    public async signedGetKey (key : string) {
        const s3Params : S3.GetObjectRequest = {
            Bucket: this.bucket,
            Key: key,
            //Expires: new Date((new Date()).getTime() + 24 * 60 * 60 * 1000) //1 day
            //ACL: 'public-read',
            //Expires: 60 ?
        };
        return new Promise((resolve : Function, reject : Function) => {
            return this.s3.getSignedUrl('getObject', s3Params, (err : Error, url : string) => {
                if (err) {
                    return reject(err);
                }
                return resolve(url);
            });
        });
    }

    /*
    readStream (to express or server)
    s3.getObject(params)
    .on('httpHeaders', function (statusCode, headers) {
        res.set('Content-Length', headers['content-length']);
        res.set('Content-Type', headers['content-type']);
        this.response.httpResponse.createUnbufferedStream()
            .pipe(res);
    })
    .send();
--------
    var fileStream = fs.createWriteStream('/path/to/file.jpg');
    var s3Stream = s3.getObject({Bucket: 'myBucket', Key: 'myImageFile.jpg'}).createReadStream();

    // Listen for errors returned by the service
    s3Stream.on('error', function(err) {
        // NoSuchKey: The specified key does not exist
        this.log.error(err);
    });

    s3Stream.pipe(fileStream).on('error', function(err) {
        // capture any errors that occur when writing data to the file
        this.log.error('File Stream:', err);
    }).on('close', function() {
        this.log.info('Done.');
    });
    */

    /**
     * Delete an object at a specific key
     *
     * @param {string} key Key for object
     *
     * @returns {boolean} True if successful
     **/

    public async delete (key : string) {
        if (!this.writeable) return false;
        const params : S3.DeleteObjectRequest = {
        	Bucket : this.bucket,
        	Key : key
        };
        return new Promise((resolve : Function, reject : Function) => {
    		return this.s3.deleteObject(params, (err : Error, data : any) => {
    			if (err) {
    				this.log.error(err);
    				return reject(err);
    			}
    			return resolve(true) //buffer
    		});
    	});
    }

    /**
     * Lists all objects with a specific prefix
     **/

    public async list (prefix : string) {
        const params : S3.ListObjectsV2Request = {
            Bucket: this.bucket,
            Prefix: prefix
        };
        return new Promise((resolve : Function, reject : Function) => {
            return this.s3.listObjectsV2(params, (err : Error, data : any) => {
                if (err) {
                    this.log.error(err);
                    return reject(err);
                }
                return resolve(data);
            });
        });
    }

    public async update () {
        if (!this.writeable) return false;
    }
}


module.exports = { Files3 };
