import 'dotenv/config';
import { createLog } from './log';
import type { Logger } from 'winston';
import { readFile, readdir, realpath } from 'fs/promises';
import { join, basename } from 'path';
import { promisify } from 'util';
import { Shell } from './shell';
import { Hashes } from './hash';
import { Files3 } from './files3'
import { envString } from './env';
import { DB } from './db';
import type { Photo } from './db';

const sizeOf = promisify(require('image-size'));

interface Metadata {
	year? : number;
	month? : number;
	day? : number;
	format?: string;
	filmstock?: string;
	location? : string;
	description? : string;
	width? : number;
	height? : number;
	original?: string;
}

class Generate {
	private log : Logger;
	private files : string[];
	private inbox : string = envString('INBOX', '~/Photos/toprocess');
	private photos : string = envString('PHOTOS', '~/Photos/processed');
	private s3 : Files3;
	private db : DB;
	
	constructor () {
		this.log = createLog('generate');
		this.log.info(`Generating site: ${new Date()}`);
		this.db = new DB();
		this.s3 = new Files3(envString('S3_BUCKET', 'mmcwilliamsphotos'), true);
		this.generate();
	}

	private async generate () {
		//check version

		//sync
		await this.checkInbox();
		//validate
	}

	private async checkInbox () {
		let inbox : string;
		let images : string[];
		let filename : string;
		let meta : Metadata;
		let dimensions : any;
		let photo : Photo;

		try {
			inbox = await realpath(this.inbox);
		} catch (err) {
			this.log.error(err);
			return;
		}

		try {
			images = await readdir(inbox); 
		} catch (err) {
			this.log.error(err);
			return;
		}

		images = images.filter((el : string) => {
			if (el.toLowerCase().indexOf('.jpg') !== -1
				|| el.toLowerCase().indexOf('.jpeg') !== -1
				|| el.toLowerCase().indexOf('.tif') !== -1
				|| el.toLowerCase().indexOf('.tiff') !== -1) {
				return true;
			}
			return false;
		});

		if (images.length === 0) {
			this.log.info(`No new images found`);
			return;
		}

		images = await Promise.all(images.map(async (el : any) : Promise<string> => {
				return await realpath(join(inbox, el));
			})
		);
		for (let image of images) {
			this.log.info(image);
			filename = basename(image);
			meta = this.parseFilename(filename);
			dimensions = await this.getImageDimensions(image);
			meta.width = dimensions.width;
			meta.height = dimensions.height;

			photo = await this.createPhoto(image, meta);

			console.dir(meta)
		}
	}

	private async img (file : string, exif : string)  {
		const cmd : string[] = ['bash', 'scripts/img.sh', file, exif];
		const shell : Shell = new Shell(cmd);
		try {
			await shell.execute();
		} catch (err) {
			this.log.error(err);
			return;
		}
		this.log.info(`Processed image file for ${file}`);
	}

	async getImageDimensions (imagePath: string): Promise<{ width: number, height: number }> {
		let dimensions : any;
		try {
			dimensions = await sizeOf(imagePath);
			return dimensions;
		} catch (err) {
			this.log.error('Error getting image dimensions:', err);
		}
	}

	private capitalize (str : string) : string {
		return (str.substring(0, 1).toUpperCase()) + str.substring(1);
	}

	private formatProperNouns (str : string) : string {
		let parts : string[] = str.split('-');
		parts = parts.map(el => this.capitalize(el));
		return parts.join(' ');
	}

	//year
	//month
	//day
	//format
	//filmstock
	//location
	//description
	//original

	//2024_12_02_35mm_Kodak-Gold-200_Somerville-MA_Walk-towards-Harvard-Square#000061280009.tif

	private parseFilename (filename : string) : Metadata {
		const halves : string[] = filename.split('#')
		const parts : string[] = halves[0].split('_');
		let meta : Metadata = {};
		for (let i = 0; i < parts.length; i++) {
			switch (i) {
				case 0 :
					meta.year = parseInt(parts[i]);
					break;
				case 1 :
					meta.month = parseInt(parts[i]);
					break;
				case 2 :
					meta.day = parseInt(parts[i]);
					break;
				case 3 : 
					meta.format = parts[i];
					break;
				case 4:
					meta.filmstock = parts[i].split('-').join(' ');
					break;
				case 5 :
					meta.location = parts[i].split('-').join(' ');
					break;
				case 6 :
					meta.description = parts[i].split('-').join(' ');
					break;
			}
		}
		meta.original = halves[1];
		return meta;
	}

	private async createPhoto (image : string, meta : Metadata) : Promise<Photo> {
		const hash : string = await Hashes.fileHash(image);
		return {
			name : basename(image),
			original: meta.original,
			hash,
			width : meta.width,
			height : meta.height,
			format : meta.format,
			filmstock : meta.filmstock,
			location : meta.location,
			discovered : Date.now(),
			created : + new Date(meta.year, meta.month, meta.day)
		}
	}
}

new Generate();